local coaf = require "coaf.core"
local ev = require "ev"
local _dbg = require "coaf.debug"
local compa = require "coaf.compatible"

local _M = {}

function _M.createNew(pipeFilter)
--====================================================
	pipeFilter = pipeFilter or "DEFAULT-PIPE"
	local PIPE_LOG = {
		defaultFilterName = pipeFilter,
		defaultFunc = _dbg.Print,
		ready = false,
		pipe_read_fd = nil,
		pipe_write_fd = nil
	}

	function PIPE_LOG.SetFilterName(filName)
		PIPE_LOG.defaultFilterName = filName or "DEFAULT-PIPE"
	end

	function PIPE_LOG.SetLogFunction(func)
		PIPE_LOG.defaultFunc = func
	end

	function PIPE_LOG.OutputLine(line)
		PIPE_LOG.defaultFunc( PIPE_LOG.defaultFilterName, line )
	end

	function PIPE_LOG.Start()
		if PIPE_LOG.ready then return true, PIPE_LOG.pipe_write_fd, PIPE_LOG.pipe_read_fd end
		local unistd = require "posix.unistd"
		local fcntl = require "posix.fcntl"
		local errno = require "posix.errno"

		local read_fd, write_fd = unistd.pipe()
		if not read_fd or not write_fd then
			return false, nil, nil, "Fail to create IO pipe for stdio monitor"
		end

		local ver = _VERSION:match("Lua (.+)$")
		ver = tonumber(ver) or 5.1

		--Non-blocking read pipe.
		local flag = fcntl.fcntl( read_fd, fcntl.F_GETFL )
		if compa.lua_version >= 5.3 then
			flag = flag | fcntl.O_NONBLOCK
		elseif compa.bit then
			flag = compa.bit.bor(flag, fcntl.O_NONBLOCK)
		end
		fcntl.fcntl( read_fd, fcntl.F_SETFL, flag )
		--TODO: This be not very good.
		--      read side: sub-process don't inherit from
		fcntl.fcntl( read_fd, fcntl.F_SETFD, 1 )

		--Non-blocking write pipe.
		flag = fcntl.fcntl( write_fd, fcntl.F_GETFL )
		if compa.lua_version >= 5.3 then
			flag = flag | fcntl.O_NONBLOCK
		elseif compa.bit then
			flag = compa.bit.bor(flag, fcntl.O_NONBLOCK)
		end
		fcntl.fcntl( write_fd, fcntl.F_SETFL, flag )

		PIPE_LOG.lineBuffer = ''

		local task = coaf.NewIO( function(task_obj)
			while true do
				local buf,e,errcode = unistd.read( read_fd, 256 )
				if not buf then
					if errcode ~= errno.EAGAIN and errcode ~= errno.EWOULDBLOCK then
						task_obj:destroy()
					end
					break
				elseif #buf == 0 then
					task_obj:destroy()
					break
				end

				local pre,nx = buf:match("(.-)[\r]?[\n](.*)$")
				if pre then
					buf = PIPE_LOG.lineBuffer .. pre	--A whole line
					PIPE_LOG.OutputLine(buf)		
					buf = nx
					pre,nx = buf:match("(.-)[\r]?[\n](.*)$")
					while pre do
						PIPE_LOG.OutputLine(pre)
						buf = nx
						pre,nx = buf:match("(.-)[\r]?[\n](.*)$")
					end
					PIPE_LOG.lineBuffer = buf
				else
					PIPE_LOG.lineBuffer = PIPE_LOG.lineBuffer .. buf
					if #PIPE_LOG.lineBuffer >= 4096 then --Too long.
						PIPE_LOG.OutputLine(PIPE_LOG.lineBuffer)
						PIPE_LOG.lineBuffer = ''
					end
				end
			end
		end, read_fd, ev.READ )
		if not task then
			return false, nil, nil, "Fail to monitor stdio"
		end
		task:start()
		PIPE_LOG.pipe_read_fd = read_fd
		PIPE_LOG.pipe_write_fd = write_fd
		PIPE_LOG.ready = true
		return true, write_fd, read_fd
	end

	return PIPE_LOG
--====================================================
end

return _M

