local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local run_daemon = require "coaf.run_daemon"
local thread = require "coaf.thread"
local socket = require "socket"
local http = require "socket.http"
local ltn12 = require "ltn12"
local cjson = require "cjson"
local pt = require "pl.pretty"
local get_prox = require 'get_prox'

pcall(require, "version")
pcall(require, "sys_setting")

local function GetSerialNo()
    if type(_G.KV) == "table" and type(_G.KV.SERIAL_HISTORIES) == "table" and #_G.KV.SERIAL_HISTORIES > 0 then
        return _G.KV.SERIAL_HISTORIES[ #_G.KV.SERIAL_HISTORIES ].serial
    else
        return "0000000000"
    end
end

local _M = {
    versionurl = "http://172.11.49.70:60000/api/firmwares/latest.json"
}

local function GetVersionInfo(api_path, deviceid, timeout)
	http.TIMEOUT = timeout or 5
	local params = {
		id = GetSerialNo()
	}
	local request_body = cjson.encode(params)
    local response_body = {}
    local res, code, response_headers = http.request{
        url = api_path ,
		method = "POST",
		headers ={
			["Content-Type"] = "application/json",
			["Content-Length"] = #request_body,
		},
        source = ltn12.source.string(request_body),
        sink = ltn12.sink.table(response_body)
    }
    
	if res and code == 200  then
        local json_str = table.concat(response_body)
        local ok, json = pcall(cjson.decode,json_str)        
        if ok and json and json.result == "ok" then
            if json and json.data then
                return json.data
            end
		end
	end
	if code == "timeout" then
		_dbg.Error("UPGARDE","timeout")
	end
	return nil
end

local function setDownload( firmwareUrl )
    _dbg.Info("UPGARDE","firmwareUrl:", firmwareUrl)
    local proxy = get_prox.GetProxy('/', 'downloader')
    if proxy then
        local r, status, err = proxy.StartDownload("firmware", firmwareUrl, "upgrade")
        proxy:Destroy()
        if not r or not status then
            _dbg.Error("UPGARDE","setDownload error", err)
            return false
        end
    else
        _dbg.Error("UPGARDE","get proxy error")
    end
    return true
end

function _M.checkUpgrade()
    local vinfo = GetVersionInfo(_M.versionurl)
    if _G.SOFTWARE_VERSION and vinfo and vinfo.version and vinfo.version ~= _G.SOFTWARE_VERSION then
        _dbg.Info("local version:", _G.SOFTWARE_VERSION, "latest version:",vinfo.version)
        _dbg.Info("UPGARDE","Need UPGARDE")

        if vinfo.downloadurl then
            _dbg.Info("UPGARDE","downloadurl:", vinfo.downloadurl)
            setDownload(vinfo.downloadurl)
        end        
    end
end

return _M