local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local cjson = require "cjson"
local pt = require "pl.pretty"
local net_status = require "net_status"
local cfglink = require "link_config"
local client = require "coaf.client"

local roll = require "coaf.log.rollfile"
local log_pipe = require "log_pipe"
local run_daemon = require "coaf.run_daemon"

_G.link_config = cfglink.Load()
_G.PIPES = _G.link_config.polydevs
-- pt.dump(_G.PIPES)

-----------------------------------------------------------
require "bonding_vpn"
local BOND = BONDING_VPN			--bonding_vpn模块暴露出来的namespace
local LOG_TO = "/var/log/bonding"
local BOND_HDR_SIZE = 24	--KCP头大小
local BOND_FEC_HDR_SIZE = 8	--FEC头大小

-----------------------------------------------------------

pcall(require, "sys_setting")

local function GetCoAFProxy( objPath, service, loc, timeout )
	local proxy
	local cli
	cli = client.New( service, loc or "local", nil, timeout or 3 ) --3 seconds timeout.
	if not cli then
		_dbg.Error( "BOND", "Fail to create connection to service: " .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "BOND", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

local function GetSerialNo()
    if type(_G.KV) == "table" and type(_G.KV.SERIAL_HISTORIES) == "table" and #_G.KV.SERIAL_HISTORIES > 0 then
        return _G.KV.SERIAL_HISTORIES[ #_G.KV.SERIAL_HISTORIES ].serial
    else
        return "0000000000"
    end
end

local function GetDevName(devname)
    if string.sub(devname,1,string.len("modem"))=="modem" then
        local nm_proxy = GetCoAFProxy( "/raw/NETWORK_MANAGER", "nm_daemon" )
        if not nm_proxy then
            return nil, "Not nm_proxy"
        end        
        local r,data = nm_proxy.GetPPPbyName(devname)
        nm_proxy:Destroy()
        if r and data then
            return data
        else
            return nil
        end
    else
        return devname
    end
end

local _M = {
	config_path = "/data/configs/client.json",
	init_thread = nil,
	running = nil,
    HTTP_TIMEOUT = 10,
    disconnect = 0,
    vpn = nil,
    conf = nil,
    Status = "aborted",
    count = 0,
    changed = false,
    inited = false,
}

local function httpapi(event, path, host,  request_body, method, timeout)
    local f = io.popen( "/data/shine/polyworker/curl.lua " .. host .. " " .. path .. " " .. request_body .. 
    " " .. method .. " " .. timeout .. " 2>&1", "r" )
    if not f then
		return 1
	else
		local ret = ""
		local l = f:read("l")
        while l do
            ret = ret .. l
			--TODO: Check message to get more details, but now ignore it.
			l = f:read("l")
		end
		local r,status,code = f:close()
		if not r or status ~= "exit" or code ~= 0 then
			--do nothing
		else
			event:Notify( "progress", { data = ret } )
        end
		return code
	end
end

local function deepcompare(t1,t2,ignore_mt)
    local ty1 = type(t1)
    local ty2 = type(t2)
    if ty1 ~= ty2 then return false end
    -- non-table types can be directly compared
    if ty1 ~= 'table' and ty2 ~= 'table' then return t1 == t2 end
    -- as well as tables which have the metamethod __eq
    local mt = getmetatable(t1)
    if not ignore_mt and mt and mt.__eq then return t1 == t2 end
    for k1,v1 in pairs(t1) do
    local v2 = t2[k1]
    if v2 == nil or not deepcompare(v1,v2) then return false end
    end
    for k2,v2 in pairs(t2) do
    local v1 = t1[k2]
    if v1 == nil or not deepcompare(v1,v2) then return false end
    end
    return true
end

local function GetConf(deviceid, authcode, timeout)
    if not timeout then
        timeout = _M.HTTP_TIMEOUT
    end
    local params = {
        id = deviceid,
        authcode = authcode,
    }
    local reqbody = cjson.encode(params)
    reqbody = string.gsub(reqbody,"\"","\\\"")
    local host = "http://" .. _G.link_config.addr .. ":" ..  _G.link_config.port
    local path = "/api/device/poly.json"

    if _M.conf_task then
        -- _M.conf_task:Destroy()
        -- _M.conf_task = nil
        -- collectgarbage()
        _dbg.Error("BOND", "conf_task is running!")
        return
    end

    local conf_task = run_daemon.Create({
        name = "getconf",
        run = function(event, path, host, reqbody, method, timeout)
            local r = httpapi( event, path, host, reqbody, method, timeout )
            os.exit(r or 0)
        end,

        args = { path, host, reqbody, "GET", timeout },

        on_terminate = function(task_obj,flag,reason,code)
            if task_obj.owner and  task_obj.owner.data then
                local ok, json = pcall(cjson.decode,task_obj.owner.data)        
                if ok and json.result == "ok" and json.data then
                    _dbg.Info("BOND", "Get poly conf Succeed!")                    
                    if _M.conf and not deepcompare(_M.conf, json.data) then
                        _dbg.Info("BOND", "Conf changed!")
                        _M.changed = true
                    end

                    if _M.changed then
                        _M.changed = false
                        coaf.ScheduleDelayedTask(function()
                            pt.dump(json.data)
                            _M.conf = json.data
                            if _M.vpn then
                                _M.Destroy()
                                coaf.Sleep(5)
                            end
                            local r,err = _M.initTUN(_M.conf)
                            if not r then
                                _dbg.Error("BOND", "initTUN error", err)
                                _M.changed = true
                            else
                                _dbg.Info("BOND", "initTUN successed!")
                                _M.inited = true
                            end
                        end)
                    end
                    _M.conf_task = nil
                    collectgarbage()
                    return nil
                else           
                    coaf.Sleep(1)   
                    return "restart"
                end
            elseif reason ~= "kill" then
                coaf.Sleep(1)
                return "restart"
            else
                _M.conf_task = nil
                collectgarbage()
                return nil
            end
        end,

        on_event = function(task_obj, event_id, msg)
            if msg and msg.data then
                task_obj.owner.data = msg.data
            end
        end
    })

    if conf_task then
        _M.conf_task = conf_task
        _M.conf_task.owner = {}
        _M.conf_task:Start()
        return true
    else        
        _dbg.Error("BOND", "Fail to start getconf task")
        return nil, "Fail to start getconf task"
    end
end

local function initlog()
    BOND.init_log()
    BOND.enable_log_console(BOND.LOG_INFO, true)
    BOND.enable_log_console(BOND.LOG_KCP, false)
    BOND.enable_log_console(BOND.LOG_FEC, false)
    BOND.enable_log_console(BOND.LOG_CRYPTO, false)
    BOND.enable_log_console(BOND.LOG_PIPE_STATE, false)
    BOND.enable_log_console(BOND.LOG_TX, false)
    BOND.enable_log_console(BOND.LOG_RX, false)
    BOND.enable_log_console(BOND.LOG_HANDSHAKE, false)
    BOND.enable_log_console(BOND.LOG_ROUTE, false)
    
    BOND.set_log_level(BOND.LOG_KCP, 0)
    BOND.set_log_level(BOND.LOG_FEC, 0)
    BOND.set_log_level(BOND.LOG_CRYPTO, 0)
    BOND.set_log_level(BOND.LOG_PIPE_STATE, 0)
    BOND.set_log_level(BOND.LOG_TX, 0)
    BOND.set_log_level(BOND.LOG_RX, 0)
    BOND.set_log_level(BOND.LOG_HANDSHAKE, 0)
    BOND.set_log_level(BOND.LOG_ROUTE, 0)    
end

--初始化聚合接口参数
local function init_pipes()
    os.execute("/bin/echo 1 > /proc/sys/net/ipv4/ip_forward")
    os.execute("/bin/echo 0 > /proc/sys/net/ipv4/conf/all/rp_filter")
end

------------------------------------------------------------
local function config_if(ifname, mtu)
    os.execute("/sbin/ifconfig " .. ifname .. " 10.0.0.1 pointopoint 10.0.0.2 mtu " .. mtu .. " up")
end

local function config_if_down(ifname)
    os.execute("/sbin/ifconfig " .. ifname .. " down")
end

-- print(GetConf("http://192.168.0.64:81/api/device/poly.json","12345678901234","QURQNRWF3P"))
function _M.initTUN(conf)
    if not conf then
        return false, "conf empty!"
    end

    initlog()   

    local bond_options = BOND.bonding_options()
    bond_options.conv_id = conf.conv_id 		
    bond_options.mtu = conf.mtu
    bond_options.send_window_size = conf.send_window_size
    bond_options.recv_window_size = conf.recv_window_size
    bond_options.nodelay = conf.nodelay
    bond_options.fast_resend = conf.fast_resend		
    bond_options.no_congestion_ctrl = conf.no_congestion_ctrl
    bond_options.update_interval = conf.update_interval

    bond_options.data_shards = conf.data_shards
    bond_options.parity_shards = conf.parity_shards
    bond_options.crypto_password = conf.crypto_password

    local vpn_mtu_size = bond_options.mtu - BOND_HDR_SIZE
    if bond_options.data_shards > 0 or bond_options.parity_shards > 0 then
        vpn_mtu_size = vpn_mtu_size - BOND_FEC_HDR_SIZE
    end
    
    if not _M.vpn then
        local vpn = BOND.create_new("/dev/net/tun")
        if not vpn then
            _dbg.Error("BOND", "Create VPN Object failed!")
            return false, "Create VPN Object failed!"
        end   
        _M.vpn = vpn

        init_pipes()

        --创建VPN网络接口（第二个参数为空字符串，由系统自动确认接口名称并返回）
        local ifname = _M.vpn:create_link(bond_options, "tun88")
        if ifname == "" then
            _dbg.Error("BOND", "Create VPN link failed!")
            _M.Destroy()
            return false, "Create VPN link failed!"
        end

        _M.ifname = ifname    
        config_if( ifname, vpn_mtu_size )
    end
    
    --客户端模式，添加管道
    local _i, item
    for _i, item in ipairs(_G.PIPES) do
        local pipeif = GetDevName(item.ifname)
        if pipeif then
            os.execute("/bin/echo 0 > /proc/sys/net/ipv4/conf/" .. pipeif .. "/rp_filter")
            local sfd = _M.vpn:add_client_pipe( item.id, conf.server_ip, conf.server_port, pipeif )
            if not sfd or sfd < 0 then
                _dbg.Error("BOND", "Add VPN client side pipe (id=", item.id, ", interface=", (pipeif or "<nil>"), ") failed!")
            else
                _dbg.Info("BOND", "add VPN client side pipe (id=", item.id, ", interface=", (pipeif or "<nil>"), ") sucess!")
            end
        else
            _dbg.Error("BOND", "pipeif:", item.ifname ,"not init yet!")
        end
    end
    return true
end

local function checkRoute()
    local f = io.popen( "/sbin/ip route", "r" )
    if f then
        local l = f:read("l")
        while l do
            -- print(l)
            local s, e = string.find(l, "default via 10.0.0.1 dev tun88  metric 50")
            if s and e then
                return true
            end
            l = f:read("l")
        end
        f:close()
    end
    return false
end

local function pingTask(thrObj, test_ip)
    local done,reach,result = net_status.Ping( test_ip )
	if thrObj.terminated then 
		_M.check_tasks = nil
		_M.Status = "aborted"
		return
	end

	if not done or not reach or result >= 100 then
        _dbg.Warn( "TUN", "unreachable" )
        -- os.execute("/sbin/ip route del ".. _G.link_config.addr .."/32 dev tun88 src 10.0.0.1 metric 100")
        os.execute("/sbin/ip route del default via 10.0.0.1 dev tun88  metric 50")
		_M.check_tasks = nil
        _M.Status = "unreachable"
		return
    end
    
    _M.check_tasks = nil
    _M.Status = "poly-ready"
    if not checkRoute() then
        _dbg.Info("BOND", "Set route to host!")
        -- os.execute("/sbin/ip route add ".. _G.link_config.addr .."/32 dev tun88 src 10.0.0.1 metric 100")
        os.execute("/sbin/ip route add default via 10.0.0.1 dev tun88  metric 50")
    end
    -- _dbg.Info("TUN",_M.Status)
    return
end

local function checkConnections()
    local ret = {}
    if _M.vpn and _M.conf then
        local conns = _M.vpn:get_pipes()
        local _i,v
        for _i,v in ipairs(conns) do
            ret[tostring(v)] = true
            local pipe_state = _M.vpn:get_pipe_state(v)
            if pipe_state and pipe_state == "DEAD" then
                _dbg.Error("BOND", "Server connection error!!! ID:", v)
                if not _M.vpn:remove_pipe(v) then
                    _dbg.Error("BOND", "remove pipe error", v)
                else
                    ret[tostring(v)] = false
                end
            end
        end

        local i, item
        for i, item in ipairs(_G.PIPES) do
            if not ret[tostring(item.id)] then
                local pipeif = GetDevName(item.ifname)
                if pipeif and _M.vpn then
                    os.execute("/bin/echo 0 > /proc/sys/net/ipv4/conf/" .. pipeif .. "/rp_filter")
                    local sfd = _M.vpn:add_client_pipe( item.id, _M.conf.server_ip, _M.conf.server_port, pipeif )
                    if not sfd or sfd < 0 then
                        _dbg.Error("BOND", "RE add VPN client side pipe (id=", item.id, ", interface=", (pipeif or "<nil>"), ") failed!")
                    else
                        _dbg.Info("BOND", "add VPN client side pipe (id=", item.id, ", interface=", (pipeif or "<nil>"), ") sucess!")
                    end
                else
                    _dbg.Error("BOND", "pipeif", item.ifname ,"not init yet!")
                end
            end
        end
    end
    -- return ret
end

function _M.Check()
    if _G.link_config.poly_enabled == 0  then
        return
    end

    if _M.inited then
        local now = os.time()
        local task = thread.New( pingTask )
        if task then
            _M.check_tasks = task
            _M.last_check_times = now
            task:Start("10.0.0.2")
        end
        checkConnections()
    end

    _M.count = _M.count + 1
    if _M.count >= 3 then
        _M.count = 0
        _M.Create()
    end
end

function _M.Get_Status()
    return _M.Status
end

function _M.Create(force)
    if _G.link_config.poly_enabled == 0 then
        -- _dbg.Info("BOND","Disable PolyService")
        if _M.vpn then
            _M.Destroy()
        end
        return false
    end

    local deviceid = GetSerialNo()
    local authcode = _G.link_config.authcode
    -- _dbg.Info("BOND","Start Create deviceid:",deviceid,"authcode:",authcode)
    GetConf(deviceid, authcode, _G.link_config.timeout)
    return true
end

function _M.Destroy()
    os.execute("/sbin/ip route del default via 10.0.0.1 dev tun88  metric 50")
    if _M.conf_task then
        _M.conf_task:Destroy()
        _M.conf_task = nil
    end

    if _M.vpn then
        _M.vpn:remove_pipe(-1) --remove all pipes-

        if _M.ifname then
            config_if_down(_M.ifname)
            _M.ifname = nil
        end
        
        _M.vpn:destroy_link()
        _M.vpn = nil
    end
    collectgarbage()
end

local function patchOldConfig()
    local file = io.open( "/data/configs/netlink.conf", "r" )
    if not file then
            return
    end
    local trunk = file:read("*a")
    file:close()
    local r
    local func
    func,r = load( "return " .. trunk )
    local d
    r, d = pcall(func)
    if func then
        if r and d then                
            if not d.polydevs then
                _G.link_config.addr = "172.11.49.195"
                _G.link_config.polydevs = {
                    { id = 1, ifname = "modem1" },
                    { id = 2, ifname = "modem2" },
                    { id = 3, ifname = "modem3" },
                }
                _G.PIPES = _G.link_config.polydevs
                _G.link_config.poly_enabled = 1
                _G.link_config:__TRIGGER(true)
            end
        end
    end
end

function _M.Init()
    patchOldConfig()
    -- pt.dump(_G.link_config)
    run_daemon.Init()
    _M.changed = true
    _M.inited = false
    _M.Create()
end

function _M.SetNetLink(addr, port, authcode, timeout, devs, poly_enabled)
    if not addr or not port or not authcode or not timeout or not devs or not poly_enabled then
        return false, "Invalid arguments"
    end

    if type(addr) ~= "string" then
        return false, "Invalid addr"
    end

    port = tonumber(port)
    if not port then
        return false, "Invalid port"
    end

    timeout = tonumber(timeout)
    if not timeout or timeout < 0 then
        return false, "Invalid timeout"
    end

    if type(devs) ~= "table" then
        devs = {
            { id = 1, ifname = "modem1" },
            { id = 2, ifname = "modem2" },
        }
    end

    poly_enabled = tonumber(poly_enabled)
    if not poly_enabled or ( poly_enabled ~= 1 and poly_enabled ~= 0) then
        return false, "Invalid enable arg"
    end

    if _G.link_config.addr ~= addr then
        _G.link_config.addr = addr
        _M.changed = true
    end

    if _G.link_config.port ~= port then
        _G.link_config.port = port
        _M.changed = true
    end

    if _G.link_config.authcode ~= authcode then
        _G.link_config.authcode = authcode
        _M.changed = true
    end

    if _G.link_config.timeout ~= timeout then
        _G.link_config.timeout = timeout
        _M.changed = true
    end

    if _G.link_config.poly_enabled ~= poly_enabled then
        _G.link_config.poly_enabled = poly_enabled
        _M.changed = true
    end

    if not deepcompare(_G.link_config.polydevs, devs) then
        _G.link_config.polydevs = devs
        _G.PIPES = _G.link_config.polydevs
        _M.changed = true
    end

    if _M.changed then
        _dbg.Info("BOND","Poly dev configure changed!")
        _M.Status = "unreachable"

        if _M.conf_task then
            _M.conf_task:Destroy()
            _M.conf_task = nil
            collectgarbage()
        end

        if _M.vpn then
            _M.Destroy()
        end

        if _G.link_config.poly_enabled == 1 then
            _M.inited = false
            _M.Create()
        else
            _dbg.Info("BOND","Poly Link Disabled!")
        end
    end

    _G.link_config:__TRIGGER(true)
    return true
end

function _M.GetNetLink()
    return _G.link_config
end

local function cutdownstr(String,End)
    if not String then
        return ""
    end
    if string.sub(String,-string.len(End))==End then
        return string.sub(String, 1, string.len(String) - string.len(End))
    else
        return String
    end
end

local function toFixed( num, digits )
    local value = tonumber(num)
    digits = digits or 0
    return string.format("%." .. tonumber(digits) .. "f", value)
end

local function byteConvert(bytes)
    if not bytes then
        return ""
    end

    local conTAB = { '', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y' }
    local exp = math.floor(math.log(bytes)/math.log(2))
    if exp < 1 then
        exp = 0
    end

    local i = math.floor(exp / 10)
    bytes = bytes / math.pow(2, 10 * i);

    if string.len(tostring(bytes)) > string.len(toFixed(bytes,2)) then
        bytes = toFixed(bytes,2);
    end
    bytes = cutdownstr(tostring(bytes),".0")
    return bytes .. conTAB[i+1]
end

-- print(byteConvert(123534532))
-- print(byteConvert(0))
-- print(byteConvert(1024))
-- print(byteConvert(1048576))


local function GetPolyStats()
    local ret = {
        link = "unknown",
        send = "unknown",
        recv = "unknown",
        rtt = "unknown",
        lostrate = "unknown",
        sendrate = "unknown",
        recvrate = "unknown",
    }

    if _M.vpn then
        local conns = _M.vpn:get_pipes()
        if conns then
            ret.link =  tostring( #conns ) .. " / " .. tostring( #_G.PIPES )
            ret.rtt = ""
            ret.send = ""
            ret.recv = ""
            ret.lostrate = ""
            ret.sendrate = ""
            ret.recvrate = ""

            local _i,conn_id
            for _i,conn_id in pairs(conns) do
                local pipe_info = _M.vpn:get_pipe_info(conn_id)
                if pipe_info then
                    --TODO 统计发送接收
                    --RTT
                    ret.rtt = ret.rtt .. byteConvert(pipe_info.srtt) .. " | "
                    --send total
                    ret.send = ret.send .. byteConvert(pipe_info.longterm_tx_bytes) .. " | "
                    ret.recv = ret.recv .. byteConvert(pipe_info.longterm_rx_bytes) .. " | "

                    ret.lostrate =  ret.lostrate .. byteConvert(pipe_info.shorterm_tx_pkg_loss_rate) .. " | "
                    ret.sendrate = ret.sendrate .. byteConvert(pipe_info.shorterm_tx_bitrate) .. " | "
                    ret.recvrate = ret.recvrate .. byteConvert(pipe_info.shorterm_rx_bitrate) .. " | "
                else
                    ret.rtt = ret.rtt .. "unknown | "
                    ret.send = ret.send .. "unknown | "
                    ret.recv = ret.recv .. "unknown | "
                    ret.lostrate =  ret.lostrate .. "unknown | "
                    ret.sendrate = ret.sendrate .. "unknown | "
                    ret.recvrate = ret.recvrate .. "unknown | "
                end
            end
            ret.rtt = cutdownstr(ret.rtt, " | ")
            ret.send = cutdownstr(ret.send, " | ")
            ret.recv = cutdownstr(ret.recv, " | ")
            ret.lostrate = cutdownstr(ret.lostrate, " | ")
            ret.sendrate = cutdownstr(ret.sendrate, " | ")
            ret.recvrate = cutdownstr(ret.recvrate, " | ")
        end
    end
    return ret
end

function _M.GetPolyStatus()
    local ret = {}
    if _G.link_config.poly_enabled == 0 then
        ret.state = "disabled"
    else
        if _M.Status == "poly-ready" then
            ret.state = "connected"
            ret.stats = GetPolyStats()
        else
            ret.state = "connecting"
        end
    end
    return ret
end

return _M

