local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local event = require "coaf.event"
--------------------------------

_G.g_event = nil

--------------------------------
local _M = {
}

--------------------------------
local function onPingNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.lossRate = msg.lossRate
end

--------------------------------
local function Ping( targetAddr, intf )
	if not targetAddr or targetAddr == "" then
		return false, false, 100
	end

	local ctx = {
		status = "wait",
		lossRate = 100
	}

	local topic = tostring(ctx)
	local ticket = g_event:Subscribe( topic, onPingNotify, ctx )
	local pid = coaf.Fork( function()
		local pfile = io.popen( "/bin/ping " .. targetAddr .. " -W 1 -c 3 -q" .. (intf and (" -I " .. intf) or ""), "r" )
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local pt, pr, lr
			while line do
				--2 packets transmitted, 2 packets received, 0% packet loss
				pt, pr, lr = line:match( "(%d+).*transmitted.*(%d+).*received.*(%d+)%%.*$" )
				if pt and pr and lr then
					break
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else
				g_event:NotifyBack( topic, { status="finished", code = code, lossRate = tonumber(lr) } )
			end
		end
		os.exit(0)
	end)

	local r,flag = coaf.WaitPID( pid )
	if not r then
		if flag == "#CoAF-KILL#" then
			os.execute( "/bin/kill " .. pid )
		end
		g_event:Unsubscribe( ticket )
		return false, false, 100
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.1)	--Wait a while
		end
		g_event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return true, ctx.code == 0, ctx.lossRate
		else
			return true, false, 100
		end
	end
end

--------------------------------
_M.Ping = Ping
--------------------------------
local function onNsLookupNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.addresses = msg.addresses
end

--------------------------------
local function NsLookup( targetHost )
	if not targetHost or targetHost == "" then
		return false, nil
	end

	local ctx = {
		status = "wait"
	}

	local topic = tostring(ctx)
	local ticket = g_event:Subscribe( topic, onNsLookupNotify, ctx )
	local pid = coaf.Fork( function()
		local pfile = io.popen( "/usr/bin/nslookup " .. targetHost )
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local found_name
			local pfx, addr, addrs
			while line do
				if not found_name then
					found_name = line:match( "^[Nn]ame:%s+(.+)$" )
				else
					addr = line:match( "^[Aa]ddress.-:%s+(.+)$" )
					if addr then
						if not addrs then
							addrs = { addr }
						else
							table.insert( addrs, addr )
						end
					end
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else
				g_event:NotifyBack( topic, { status="finished", code = code, addresses = addrs } )
			end
		end
		os.exit(0)
	end)

	local r,flag = coaf.WaitPID( pid )
	if not r then
		if flag == "#CoAF-KILL#" then
			os.execute( "/bin/kill " .. pid )
		end
		g_event:Unsubscribe( ticket )
		return false, nil
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.1)	--Wait a while
		end
		g_event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return ctx.code == 0, ctx.addresses
		else
			return false, nil
		end
	end
end
--------------------------------
_M.NsLookup = NsLookup
--------------------------------

function _M.Init()
	if not _G.g_event then
		_G.g_event = event.createNew( "polyworker", true ) -- Don't start slave thread, since we never use it.
	end
end
--------------------------------
return _M

