#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"


-- Change current dir to app location
coaf.ChDir(coaf.GetAppDir(arg))

local socket = require "cosocket.core"

socket.default_close_on_exec = true

local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
local service = require "coaf.service"
local roll = require "coaf.log.rollfile"

local withLog = false
local _x
for _x=1, #arg do
	if arg[_x] == "--log" then
		withLog = true
		break
	end
end

if withLog then
	-- ** initialize the log file **
	local roll = require "coaf.log.rollfile"
	os.execute( "/bin/mkdir -p /data/log/" )
	local logF = roll.Create("/data/log/polyworker.log")
	if not logF or not logF:init() then
		_dbg.Error("POLY", "Fail to open log file: /data/log/polyworker.log, so the log messages should be output to console.")
		withLog = false
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
	-- ** log **
end

coaf.Init( "polyworker" )

_G.print = function(...)
	_dbg.Info1( "POLY", ... )
end

if withLog then
	local stdio = require "coaf.log.stdio"
	stdio.SetDefaultFilterName( "polyworker" )
	stdio.Start()
end

-- CoAF serivce initialization.
service.InitSocket()
client.InitSocket( socket, adapter )

----------------------------------------

local NET_STATUS = require "net_status"
local multitun = require "mod_multitun"
-- local mod_autoupgrade = require "mod_autoupgrade"

----------------------------------------
local taskrunning = false
local function periodTask()
	if not taskrunning then
		taskrunning = true
		multitun.Check() 
		-- mod_autoupgrade.checkUpgrade()
		taskrunning = false
	end
end

----------------------------------------
local function Main()
	NET_STATUS.Init()
	multitun.Init()

	local cosvc_NM = service.New( "polyworker", "unix:/var/run/polyworker.sock" )
	if cosvc_NM then
		_dbg.Info( "POLY", "CoService of 'polyworker' is created." )
		cosvc_NM:CreateCoServerProxy()
		cosvc_NM:ExposeObject( "/poly", multitun )
	else
		_dbg.Error( "POLY", "Fail created CoService!" )
	end

	local period_timer = thread.NewTimer( periodTask, 1, 10 )
	if period_timer then
		period_timer:start()
	end
end
----------------------------------------

coaf.Start(Main)
multitun.Destroy()
_dbg.Info( "POLY", "Polywoker service is terminated." )
